"""Tests for epoch.py"""

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from timeflux.core.branch import Branch
from timeflux.core.exceptions import WorkerInterrupt
from timeflux.helpers import testing as helpers
from timeflux.nodes.epoch import Epoch

data = helpers.DummyData()

rate = 10
before = .2
after = .6
node = Epoch(event_trigger='test', before=before, after=after)


def test_no_epoch():
    # Send enough data for an epoch, but no event
    node.i.data = data.next(20)
    node.update()
    assert node.o.data == None


def test_send_event_short_data():
    # Send an event trigger, but not enough data to capture the epoch in one round
    node.clear()
    node.i.data = data.next(5)
    time = node.i.data.index[1]  # Sync event to second sample
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    assert node.o.data == None


def test_receive_epoch():
    # Send enough data to complete an epoch
    node.clear()
    node.i.data = data.next()
    node.update()
    expected_meta = {
        'epoch': {
            'onset': pd.Timestamp('2018-01-01 00:00:02.096394939'),
            'context': 'foobar',
            'before': before,
            'after': after
        }
    }
    expected_data = pd.DataFrame(
        [
            [0.176528, 0.220486, 0.186438, 0.779584, 0.350125],
            [0.057843, 0.969103, 0.883786, 0.927752, 0.994908],
            [0.173895, 0.396242, 0.758238, 0.696021, 0.153896],
            [0.815833, 0.224441, 0.223818, 0.536974, 0.592940],
            [0.580086, 0.091487, 0.877461, 0.265600, 0.129515],
            [0.888748, 0.955651, 0.862128, 0.809516, 0.655242],
            [0.550857, 0.086987, 0.408453, 0.372689, 0.259754],
            [0.723420, 0.495876, 0.081046, 0.220183, 0.683259]
        ],
        [
            pd.Timestamp('2018-01-01 00:00:01.897912291'),
            pd.Timestamp('2018-01-01 00:00:02.001118529'),
            pd.Timestamp('2018-01-01 00:00:02.096394939'),
            pd.Timestamp('2018-01-01 00:00:02.197921447'),
            pd.Timestamp('2018-01-01 00:00:02.298663618'),
            pd.Timestamp('2018-01-01 00:00:02.399560700'),
            pd.Timestamp('2018-01-01 00:00:02.502851760'),
            pd.Timestamp('2018-01-01 00:00:02.596996738')
        ]
    )
    assert node.o.meta == expected_meta
    pd.testing.assert_frame_equal(node.o.data, expected_data)


def test_invalid_event():
    # Make sure no epoch is fetched on an invalid trigger
    node.clear()
    node.i.data = data.next()
    time = node.i.data.index[0]
    event = pd.DataFrame([['invalid', 'foobar']], [time], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert node.o.data == None


def test_multiple_events():
    # Make sure the correct onset is caught
    node.clear()
    node.i.data = data.next()
    event = pd.DataFrame(
        [
            ['invalid_1', 'foo'],
            ['test', 'bar'],
            ['invalid_2', 'baz']
        ], [
            node.i.data.index[0],
            node.i.data.index[1],
            node.i.data.index[2]
        ], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert node.o.meta['epoch']['onset'] == pd.Timestamp('2018-01-01 00:00:04.598117111')


def test_multiple_triggers():
    # Mutliple epochs should be returned
    node.clear()
    node.i.data = data.next()
    event = pd.DataFrame(
        [
            ['test', 'foo'],
            ['test', 'bar'],
        ], [
            node.i.data.index[0],
            node.i.data.index[1],
        ], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert node.o_0.meta['epoch']['context'] == 'foo'
    assert node.o_1.meta['epoch']['context'] == 'bar'


def test_unsynced_event():
    # The epoch must be fetched even if the trigger time does not match exactly
    node.clear()
    node.i.data = data.next(20)
    time = pd.Timestamp('2018-01-01 00:00:07.100')
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    expected_indices = pd.DatetimeIndex([
        pd.Timestamp('2018-01-01T00:00:06.904868869'),
        pd.Timestamp('2018-01-01T00:00:07.002722448'),
        pd.Timestamp('2018-01-01T00:00:07.096987157'),
        pd.Timestamp('2018-01-01T00:00:07.195055221'),
        pd.Timestamp('2018-01-01T00:00:07.303154614'),
        pd.Timestamp('2018-01-01T00:00:07.402068573'),
        pd.Timestamp('2018-01-01T00:00:07.502290072'),
        pd.Timestamp('2018-01-01T00:00:07.602712703'),
        pd.Timestamp('2018-01-01T00:00:07.695740447')
    ])
    pd.testing.assert_index_equal(node.o.data.index, expected_indices)


def test_receives_event_before_data():
    # receives event before data
    # NB: this use case can also happen if before<0
    before = 0
    data.reset()
    node = Epoch(event_trigger='test', before=before, after=.6)
    node.clear()
    node.i.data = data.next(5)
    time = data._data.index[6]  # Sync event right after last sample
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    node.i_events.data = None
    node.i.data = data.next(10)
    node.update()
    expected_data = pd.DataFrame(
        [
            [0.037008, 0.59627, 0.230009, 0.120567, 0.076953],
            [0.696289, 0.339875, 0.724767, 0.065356, 0.31529],
            [0.539491, 0.790723, 0.318753, 0.625891, 0.885978],
            [0.615863, 0.232959, 0.024401, 0.870099, 0.021269],
            [0.874702, 0.528937, 0.939068, 0.798783, 0.997934],
            [0.350712, 0.767188, 0.401931, 0.479876, 0.627505]
        ],
        [
            pd.Timestamp('2018-01-01 00:00:00.595580836'),
            pd.Timestamp('2018-01-01 00:00:00.703661761'),
            pd.Timestamp('2018-01-01 00:00:00.801011149'),
            pd.Timestamp('2018-01-01 00:00:00.902080726'),
            pd.Timestamp('2018-01-01 00:00:00.995205845'),
            pd.Timestamp('2018-01-01 00:00:01.104699099'),
        ]
    )
    expected_meta = {'epoch': {'onset': pd.Timestamp('2018-01-01 00:00:00.595580836'),
                               'context': 'foobar',
                               'before': before,
                               'after': after}}
    assert node.o.meta == expected_meta
    pd.testing.assert_frame_equal(node.o.data, expected_data)


def test_empty_epoch():
    # no data received in the epoch
    before = 1
    after = -1
    data = helpers.DummyData(rate=.1)
    node = Epoch(event_trigger='test', before=before, after=after)
    node.clear()
    node.i.data = data.next(5)
    time = pd.Timestamp("2018-01-01 00:00:10.450714306")  # Sync event to second sample
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    expected_meta = {'epoch': {'onset': pd.Timestamp('2018-01-01 00:00:10.450714306'),
                               'context': 'foobar',
                               'before': before,
                               'after': after}
                     }
    assert node.o.meta == expected_meta
    assert node.o.data.empty


def test_to_xarray():
    """ Test the epoch followed by a conversion to xarray.
    """

    graph = {
        'nodes': [
            {
                'id': 'epoch',
                'module': 'timeflux.nodes.epoch',
                'class': 'Epoch',
                'params': {
                    'before': before,
                    'after': after,
                    'event_trigger': 'test'

                }
            },
            {
                'id': 'to_xarray',
                'module': 'timeflux.nodes.epoch',
                'class': 'ToXArray',
                'params': {
                    'reporting': 'error'
                }
            }
        ],
        'edges': [
            {
                'source': 'epoch:*',
                'target': 'to_xarray'
            }

        ]
    }
    # test with data that has no jitter
    data = helpers.DummyData(jitter=0, rate=rate)
    converted_epoch = Branch(graph)
    converted_epoch.set_port('epoch', port_id='i', data=data.next(20))
    event = pd.DataFrame(
        [
            ['test', 'foo'],
            ['test', 'bar'],
        ], [
            pd.Timestamp('2018-01-01 00:00:00.300000'),
            pd.Timestamp('2018-01-01 00:00:00.400000')
        ], columns=['label', 'data'])
    converted_epoch.set_port('epoch', port_id='i_events', data=event)
    converted_epoch.update()

    expected_meta = {'epochs_context': ['foo', 'bar'], 'rate': 10,
                     'epochs_onset': [pd.Timestamp('2018-01-01 00:00:00.300000'),
                                      pd.Timestamp('2018-01-01 00:00:00.400000')]}
    assert converted_epoch.get_port('to_xarray').meta == expected_meta

    expected_data = xr.DataArray(data=np.array([[[0.658783, 0.692277, 0.849196, 0.249668, 0.489425],
                                                 [0.221209, 0.987668, 0.944059, 0.039427, 0.705575],
                                                 [0.925248, 0.180575, 0.567945, 0.915488, 0.033946],
                                                 [0.69742, 0.297349, 0.924396, 0.971058, 0.944266],
                                                 [0.474214, 0.862043, 0.844549, 0.3191, 0.828915],
                                                 [0.037008, 0.59627, 0.230009, 0.120567, 0.076953],
                                                 [0.696289, 0.339875, 0.724767, 0.065356, 0.31529],
                                                 [0.539491, 0.790723, 0.318753, 0.625891, 0.885978],
                                                 [0.615863, 0.232959, 0.024401, 0.870099, 0.021269]],
                                                [[0.221209, 0.987668, 0.944059, 0.039427, 0.705575],
                                                 [0.925248, 0.180575, 0.567945, 0.915488, 0.033946],
                                                 [0.69742, 0.297349, 0.924396, 0.971058, 0.944266],
                                                 [0.474214, 0.862043, 0.844549, 0.3191, 0.828915],
                                                 [0.037008, 0.59627, 0.230009, 0.120567, 0.076953],
                                                 [0.696289, 0.339875, 0.724767, 0.065356, 0.31529],
                                                 [0.539491, 0.790723, 0.318753, 0.625891, 0.885978],
                                                 [0.615863, 0.232959, 0.024401, 0.870099, 0.021269],
                                                 [0.874702, 0.528937, 0.939068, 0.798783, 0.997934]]]),
                                 dims=('epoch', 'time', 'space'),
                                 coords=([0, 1],
                                         pd.TimedeltaIndex(np.arange(-before, after + 1 / rate, 1 / rate), unit='s'),
                                         [0, 1, 2, 3, 4]))
    xr.testing.assert_equal(converted_epoch.get_port('to_xarray').data, expected_data)

    # now test with jittered data, so that the some epoch have invalid number of samples
    data = helpers.DummyData(jitter=.2, rate=rate)
    converted_epoch = Branch(graph)
    converted_epoch.set_port('epoch', port_id='i', data=data.next(20))
    event = pd.DataFrame(
        [
            ['test', 'foo'],
            ['test', 'bar'],
        ], [
            pd.Timestamp('2018-01-01 00:00:00.300000'),
            pd.Timestamp('2018-01-01 00:00:00.400000')
        ], columns=['label', 'data'])
    converted_epoch.set_port('epoch', port_id='i_events', data=event)

    with pytest.raises(WorkerInterrupt):
        converted_epoch.update()



"""timeflux.core.message: serialize and unserialize dataframes"""

import pickle
import pandas as pd

# import pyarrow as pa


def pickle_serialize(message):
    topic = message[0]
    data = message[1]
    meta = message[2]
    return [
        topic,
        pickle.dumps(data, pickle.HIGHEST_PROTOCOL),
        pickle.dumps(meta, pickle.HIGHEST_PROTOCOL),
    ]


def pickle_deserialize(message):
    topic = message[0].decode("utf-8")
    data = message[1]
    meta = message[2]
    return [topic, pickle.loads(data), pickle.loads(meta)]


def msgpack_serialize(message):
    # TODO: handle meta and cases where data is None
    topic = message[0]
    data = message[1]
    return [topic, data.to_msgpack()]


def msgpack_deserialize(message):
    # TODO: handle meta and cases where data is None
    topic = message[0].decode("utf-8")
    data = message[1]
    return [topic, pd.read_msgpack(data)]


# def arrow_serialize(message):
#     topic = message[0].decode('utf-8')
#     df = message[1]
#     return [topic, pa.serialize(df).to_buffer()]

# def arrow_deserialize(message):
#     topic = message[0]
#     data = message[1]
#     return [topic, pa.deserialize(data)]



"""Lab streaming layer nodes

    The lab streaming layer provides a set of functions to make instrument data
    accessible in real time within a lab network. From there, streams can be
    picked up by recording programs, viewing programs or custom experiment
    applications that access data streams in real time.

"""

import pandas as pd
import numpy as np
import uuid
from pylsl import (
    StreamInfo,
    StreamOutlet,
    StreamInlet,
    resolve_stream,
    resolve_byprop,
    pylsl,
)
from time import time
from timeflux.core.node import Node


class Send(Node):

    """Send to a LSL stream.

    Attributes:
        i (Port): Default data input, expects DataFrame.

    Args:
        name (string): The name of the stream.
        type (string): The content type of the stream, .
        format (string): The format type for each channel. Currently, only ``double64`` and ``string`` are supported.
        rate (float): The nominal sampling rate. Set to ``0.0`` to indicate a variable sampling rate.
        source (string, None): The unique identifier for the stream. If ``None``, it will be auto-generated.

    Example:
        .. literalinclude:: /../examples/lsl.yaml
           :language: yaml

    """

    _dtypes = {"double64": np.number, "string": np.object}

    def __init__(self, name, type="Signal", format="double64", rate=0.0, source=None):
        if not source:
            source = str(uuid.uuid4())
        self._name = name
        self._type = type
        self._format = format
        self._rate = rate
        self._source = source
        self._outlet = None

    def update(self):
        if isinstance(self.i.data, pd.core.frame.DataFrame):
            if not self._outlet:
                labels = list(
                    self.i.data.select_dtypes(include=[self._dtypes[self._format]])
                )
                info = StreamInfo(
                    self._name,
                    self._type,
                    len(labels),
                    self._rate,
                    self._format,
                    self._source,
                )
                channels = info.desc().append_child("channels")
                for label in labels:
                    if not isinstance("string", type(label)):
                        label = str(label)
                    channels.append_child("channel").append_child_value("label", label)
                self._outlet = StreamOutlet(info)
            values = self.i.data.select_dtypes(
                include=[self._dtypes[self._format]]
            ).values
            stamps = self.i.data.index.values.astype(np.float64)
            for row, stamp in zip(values, stamps):
                self._outlet.push_sample(row, stamp)


class Receive(Node):

    """Receive from a LSL stream.

    Attributes:
        o (Port): Default output, provides DataFrame and meta.

    Args:
        prop (string): The property to look for during stream resolution (e.g., ``name``, ``type``, ``source_id``).
        value (string): The value that the property should have (e.g., ``EEG`` for the type property).
        timeout (float): The resolution timeout, in seconds.
        unit (string): Unit of the timestamps (e.g., ``s``, ``ms``, ``us``, ``ns``). The LSL library uses seconds by default. Timeflux uses nanoseconds by default.
        sync (string, None): The method used to synchronize timestamps. Use ``local`` if you receive the stream from another application on the same computer. Use ``network`` if you receive from another computer. Use ``None`` if you receive from a Timeflux instance on the same computer.
        channels (list, None): Override the channel names. If ``None``, the names defined in the LSL stream will be used.
        max_samples (int): The maximum number of samples to return per call.

    Example:
        .. literalinclude:: /../examples/lsl_multiple.yaml
           :language: yaml

    """

    def __init__(
        self,
        name=None,
        prop="name",
        value=None,
        timeout=1.0,
        unit="s",
        offset_correction=False,
        sync="local",
        channels=None,
        max_samples=1024,
    ):
        if name:
            self.logger.warning(
                'The "name" parameter is deprecated. Use "prop" and "value" instead.'
            )
            value = name
            prop = "name"
        if not value:
            raise ValueError("Please specify a stream name or a property and value.")
        if offset_correction:
            self.logger.warning(
                'The "offset_correction" parameter is deprecated. Use "sync" instead.'
            )
            sync = "local"
        self._prop = prop
        self._value = value
        self._inlet = None
        self._labels = None
        self._unit = unit
        self._sync = sync
        self._channels = channels
        self._timeout = timeout
        self._max_samples = max_samples
        self._offset = time() - pylsl.local_clock()

    def update(self):
        if not self._inlet:
            self.logger.debug(f"Resolving stream with {self._prop} {self._value}")
            streams = resolve_byprop(self._prop, self._value, timeout=self._timeout)
            if not streams:
                return
            self.logger.debug("Stream acquired")
            self._inlet = StreamInlet(streams[0])
            info = self._inlet.info()
            self._meta = {
                "name": info.name(),
                "type": info.type(),
                "rate": info.nominal_srate(),
                "info": str(info.as_xml()).replace("\n", "").replace("\t", ""),
            }
            if isinstance(self._channels, list):
                self._labels = self._channels
            else:
                description = info.desc()
                channel = description.child("channels").first_child()
                self._labels = [channel.child_value("label")]
                for _ in range(info.channel_count() - 1):
                    channel = channel.next_sibling()
                    self._labels.append(channel.child_value("label"))
        if self._inlet:
            values, stamps = self._inlet.pull_chunk(max_samples=self._max_samples)
            if stamps:
                stamps = np.array(stamps)
                if self._sync == "local":
                    stamps += self._offset
                elif self._sync == "network":
                    stamps = stamps + self._inlet.time_correction() + self._offset
                stamps = pd.to_datetime(stamps, format=None, unit=self._unit)
            self.o.set(values, stamps, self._labels, self._meta)



"""timeflux.nodes.hdf5: HDF5 nodes"""

import pandas as pd
import timeflux.helpers.clock as clock
import sys
import os
import time
from timeflux.core.exceptions import WorkerInterrupt
from timeflux.core.node import Node

# Ignore the "object name is not a valid Python identifier" message
import warnings
from tables.exceptions import NaturalNameWarning

warnings.simplefilter("ignore", NaturalNameWarning)


class Replay(Node):
    """Replay a HDF5 file."""

    def __init__(self, filename, keys, speed=1, timespan=None, resync=True, start=0):
        """
        Initialize.

        Parameters
        ----------
        filename : string
            The path to the HDF5 file.
        keys: list
            The list of keys to replay.
        speed: float
            The speed at which the data must be replayed. 1 means real-time.
            Default: 1
        timespan: float
            The timespan of each chunk, in seconds.
            If not None, will take precedence over the `speed` parameter
            Default: None
        resync: boolean
            If False, timestamps will not be resync'ed to current time
            Default: True
        start: float
            Start directly at the given time offset, in seconds
            Default: 0
        """

        # Load store
        try:
            self._store = pd.HDFStore(self._find_path(filename), mode="r")
        except IOError as e:
            raise WorkerInterrupt(e)

        # Init
        self._sources = {}
        self._start = pd.Timestamp.max
        self._stop = pd.Timestamp.min
        self._speed = speed
        self._timespan = None if not timespan else pd.Timedelta(f"{timespan}s")
        self._resync = resync

        for key in keys:
            try:
                # Check format
                if not self._store.get_storer(key).is_table:
                    self.logger.warning("%s: Fixed format. Will be skipped.", key)
                    continue
                # Get first index
                first = self._store.select(key, start=0, stop=1).index[0]
                # Get last index
                nrows = self._store.get_storer(key).nrows
                last = self._store.select(key, start=nrows - 1, stop=nrows).index[0]
                # Check index type
                if type(first) != pd.Timestamp:
                    self.logger.warning("%s: Invalid index. Will be skipped.", key)
                    continue
                # Find lowest and highest indices across stores
                if first < self._start:
                    self._start = first
                if last > self._stop:
                    self._stop = last
                # Extract meta
                if self._store.get_node(key)._v_attrs.__contains__("meta"):
                    meta = self._store.get_node(key)._v_attrs["meta"]
                else:
                    meta = {}
                # Set output port name, port will be created dynamically
                name = "o" + key.replace("/", "_")
                # Update sources
                self._sources[key] = {
                    "start": first,
                    "stop": last,
                    "nrows": nrows,
                    "name": name,
                    "meta": meta,
                }
            except KeyError:
                self.logger.warning("%s: Key not found.", key)

        # Current time
        now = clock.now()

        # Starting timestamp
        self._start += pd.Timedelta(f"{start}s")

        # Time offset
        self._offset = pd.Timestamp(now) - self._start

        # Current query time
        self._current = self._start

        # Last update
        self._last = now

    def update(self):

        if self._current > self._stop:
            raise WorkerInterrupt("No more data.")

        min = self._current

        if self._timespan:
            max = min + self._timespan
        else:
            now = clock.now()
            ellapsed = now - self._last
            max = min + ellapsed * self._speed
            self._last = now

        for key, source in self._sources.items():

            # Select data
            data = self._store.select(key, "index >= min & index < max")

            # Add offset
            if self._resync:
                data.index += self._offset

            # Update port
            getattr(self, source["name"]).data = data
            getattr(self, source["name"]).meta = source["meta"]

        self._current = max

    def terminate(self):
        self._store.close()

    def _find_path(self, path):
        path = os.path.normpath(path)
        if os.path.isabs(path):
            if os.path.isfile(path):
                return path
        else:
            for base in sys.path:
                full_path = os.path.join(base, path)
                if os.path.isfile(full_path):
                    return full_path
        raise WorkerLoadError(f"File `{path}` could not be found in the search path.")


class Save(Node):
    """Save to HDF5."""

    def __init__(
        self, filename=None, path="/tmp", complib="zlib", complevel=3, min_itemsize=None
    ):
        """
        Initialize.

        Parameters
        ----------
        filename: string
            Name of the file (inside the path set by parameter). If not set,
            an auto-generated filename is used.
        path : string
            The directory where the HDF5 file will be written.
            Default: "/tmp"
        complib : string
            The compression lib to be used.
            see: https://www.pytables.org/usersguide/libref/helper_classes.html
            Default: "zlib"
        complevel : int
            The compression level. A value of 0 disables compression.
            Default: 9
            see: https://www.pytables.org/usersguide/libref/helper_classes.html
        min_itemsize : int
            The string columns size
            Default: None
            see: https://pandas.pydata.org/pandas-docs/stable/generated/pandas.HDFStore.append.html
            see: http://pandas.pydata.org/pandas-docs/stable/io.html#string-columns

        """
        os.makedirs(path, exist_ok=True)
        if filename is None:
            filename = os.path.join(
                path, time.strftime("%Y%m%d-%H%M%S.hdf5", time.gmtime())
            )
        else:
            filename = os.path.join(path, filename)
        self.logger.info("Saving to %s", filename)
        self._store = pd.HDFStore(filename, complib=complib, complevel=complevel)
        self.min_itemsize = min_itemsize

    def update(self):
        if self.ports is not None:
            for name, port in self.ports.items():
                if not name.startswith("i"):
                    continue
                key = "/" + name[2:].replace("_", "/")
                if port.data is not None:
                    self._store.append(key, port.data, min_itemsize=self.min_itemsize)
                if port.meta is not None and port.meta:
                    # Note: not none and not an empty dict, because this operation
                    #       overwrites previous metadata and an empty dict would
                    #       just remove any previous change
                    node = self._store.get_node(key)
                    if node:
                        self._store.get_node(key)._v_attrs["meta"] = port.meta

    def terminate(self):
        try:
            self._store.close()
        except Exception:
            # Just in case
            pass



"""Arbitrary operations on DataFrames"""

from timeflux.core.node import Node
from importlib import import_module
import pandas as pd


class ApplyMethod(Node):
    """Apply a function along an axis of the DataFrame.

    This node applies a function along an axis of the DataFrame.
    Objects passed to the function are Series objects whose index is
    either the DataFrame's index (``axis`` = `0`) or the DataFrame's columns
    (``axis`` = `1`).

    Attributes:
        i (Port): default data input, expects DataFrame.
        o (Port): default output, provides DataFrame.

    Args:
       func (func): custom function specified directely that takes as input a n_array (eg. lambda x: x+1). Default: None.
       method (str): name of the module to import, in which the method is defined. eg. `numpy.mean`.
       apply_mode (str`): {`universal`, `reduce`, `expand` }. Default: `universal`.
             -  `universal` if function is a transformation from n_array to n_array
             -  `reduce` if function is a transformation from n_array to scalar
             -  `expand` if function is a transformation from n_array to nk_array [not yet implemented]
       axis (int) : if 0, the transformation is applied to columns, if 1 to rows. Default: `0`.
       closed (str) : {`left`, `right`, `center`}: timestamp to transfer in the output, only when method_type is "reduce" and axis = 0, in which case, the output port's lenght is 1. Default: `right`.
       kwargs:  additional keyword arguments to pass as keywords arguments to `func`.

    Notes:

        Note that the passed function will receive ndarray objects for performance purposes.
        For universal functions, ie. transformation from n_array to n_array, input and output ports have the same size.
        For reducing function, ie. from n_array to scalar, output ports's index is set to first (if ``closed`` = `left`), last (if ``closed`` = `right`), or middle (if ``closed`` = `center`)


    .. todo::
        Allow expanding functions such as n_array to nk_array (with XArray usage)

    Example:

        Universal function: in this example, we apply `numpy.sqrt` to each value of the data. Shapes of input and output data are the same.

        * ``method`` = `numpy.sqrt`
        * ``method_type`` = `universal`

        If data in input port is ``i`` is: ::

                                               0
            2018-10-25 07:33:41.871131         9.0
            2018-10-25 07:33:41.873084         16.0
            2018-10-25 07:33:41.875037         1.0
            2018-10-25 07:33:41.876990         4.0



        It returns the squared root of the data on port ``o``: ::

                                               0
            2018-10-25 07:33:41.871131         3.0
            2018-10-25 07:33:41.873084         4.0
            2018-10-25 07:33:41.875037         1.0
            2018-10-25 07:33:41.876990         2.0

    Example:

        Reducing function: in this example, we apply `numpy.sum` to each value of the data. Shapes of input and output data are not the same.
        We set:

        * ``method`` = `numpy.sum`
        * ``method_type`` = `reduce`
        * ``axis`` = `0`
        * ``closed`` = `right`

        If data in input port is ``i`` is: ::

                                               0        1
            2018-10-25 07:33:41.871131         9.0      10.0
            2018-10-25 07:33:41.873084         16.0     2.0
            2018-10-25 07:33:41.875037         1.0      5.0
            2018-10-25 07:33:41.876990         4.0      2.0



        It returns the sum amongst row axis on port ``o``: ::

                                               0        1
            2018-10-25 07:33:41.876990         30.0     19.0

    References:

        See the documentation of `pandas.apply <https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.apply.html>`_ .


    """

    def __init__(
        self,
        method,
        apply_mode="universal",
        axis=0,
        closed="right",
        func=None,
        **kwargs,
    ):

        self._axis = axis
        self._closed = closed
        self._kwargs = kwargs
        self._apply_mode = apply_mode

        if func is not None:
            self._func = func
        else:
            module_name, function_name = method.rsplit(".", 1)

            try:
                module = import_module(module_name)
            except ImportError:
                raise ImportError(f"Could not import module {module_name}")
            try:
                self._func = getattr(module, function_name)
            except AttributeError:
                raise ValueError(
                    f"Module {module_name} has no function {function_name}"
                )

            if not callable(self._func):
                raise ValueError(f"Could not call the method {self._methode_name}")

            self._kwargs.update({"raw": True, "axis": axis})
            if self._apply_mode == "reduce":
                self._kwargs["result_type"] = "reduce"

    def update(self):

        if not self.i.ready():
            return

        self.o.meta = self.i.meta
        self.o.data = self.i.data.apply(func=self._func, **self._kwargs)
        if self._apply_mode == "reduce":
            if self._axis == 0:
                if self._closed == "right":
                    index_to_keep = self.i.data.index[-1]
                elif self._closed == "left":
                    index_to_keep = self.i.data.index[0]
                else:  # self._closed == 'middle':
                    index_to_keep = self.i.data.index[len(self.i.data) // 2]
                self.o.data = pd.DataFrame(self.o.data, columns=[index_to_keep]).T
            else:  # self._axis == 1:
                self.o.data = self.o.data.to_frame()



"""Generate random events"""

import copy
import datetime
import json
import random
import string

import numpy as np

from timeflux.core.node import Node


class Events(Node):
    def __init__(
        self,
        rows_min=1,
        rows_max=10,
        string_min=3,
        string_max=12,
        items_min=0,
        items_max=5,
        seed=None,
    ):
        """Return random integers from value_min to value_max (inclusive)"""
        self._rows_min = rows_min
        self._rows_max = rows_max
        self._string_min = string_min
        self._string_max = string_max
        self._items_min = items_min
        self._items_max = items_max
        random.seed(seed)

    def random_string(self, length):
        return "".join(random.choice(string.ascii_letters) for m in range(length))

    def update(self):
        rows = []
        for i in range(random.randint(self._rows_min, self._rows_max)):
            row = []
            row.append(
                self.random_string(random.randint(self._string_min, self._string_max))
            )
            data = {}
            for j in range(random.randint(self._items_min, self._items_max)):
                key = self.random_string(
                    random.randint(self._string_min, self._string_max)
                )
                value = self.random_string(
                    random.randint(self._string_min, self._string_max)
                )
                data[key] = value
            row.append(json.dumps(data))
            rows.append(row)
        self.o.set(rows, names=["label", "data"])


class Periodic(Node):
    """Node that sends events at a regular interval.

    This node sends regular events after the first time the update method is
    called. If the update method is called at time `t`, and this node has a
    interval `ti` and phase `ts`, then the first event will be at `t + ts`.
    Then there will be one event at `t + ts + k ti` where `k` is 1, 2, ...

    Args:
        label (str): Event name that will be generated by this node.
        data (dict): Dictionary sent in each event.
        interval (dict): Dictionary with keyword arguments passed to
            `datetime.timedelta` to define the time interval between events.
            This can be seconds, milliseconds, microseconds, etc.
        phase (dict): Dictionary with keyword arguments passed to
            `datetime.timedelta` to define a phase for the stimulations. The
            first stimulation will happen after this time delta is observed.
            If not set, the phase will be as the interval.

    Attributes:
        o (Port): Default output, provides a pandas.DataFrame with events.

    Examples:

        The following YAML can be used to generate events every half second
        but only after 5 seconds have elapsed

        .. code-block:: yaml

           graphs:
              - nodes:
                - id: clock
                  module: timeflux.nodes.events
                  class: Periodic
                  params:
                    label: my-event-label
                    interval:
                      milliseconds: 500
                    phase:
                      seconds: 5

                - id: display
                  module: timeflux.nodes.debug
                  class: Display

                rate: 20

                edges:
                  - source: clock
                    target: display
    """

    def __init__(self, label="clock", data=None, interval=None, phase=None):
        super().__init__()

        interval = interval or {}
        phase = phase or interval
        delta_interval = datetime.timedelta(**interval)
        delta_phase = datetime.timedelta(**phase)
        if delta_interval.total_seconds() <= 0:
            raise ValueError("Periodic must have positive interval")
        if delta_phase.total_seconds() <= 0:
            raise ValueError("Periodic must have positive phase")

        self._period = np.timedelta64(delta_interval)
        self._phase = np.timedelta64(delta_phase)
        self._next_timestamp = None
        self._label = label
        self._data = data

    def update(self):
        now = np.datetime64(datetime.datetime.now())
        if self._next_timestamp is None:
            self._next_timestamp = now + self._phase
            self.logger.debug(
                "Periodic will start sending events at %s", self._next_timestamp
            )

        data = []
        times = []
        while self._next_timestamp < now:
            content = copy.deepcopy(self._data)
            data.append([self._label, content])
            times.append(self._next_timestamp)
            self._next_timestamp += self._period

        if data:
            self.logger.debug("Sending %d events", len(data))
            self.o.set(data, names=["label", "data"], timestamps=times)

    def _set_next(self, reference):
        self._next_timestamp = reference + self._period



"""timeflux.core.graph: handle graphs"""

import networkx as nx
from timeflux.core.exceptions import GraphDuplicateNode, GraphUndefinedNode


class Graph:

    """Graph helpers"""

    def __init__(self, graph):
        """
        Initialize graph.

        Parameters
        ----------
        graph : dict
          The graph, which structure can be found in `test.yaml`. Must be an acyclic directed graph.
          Multiple edges between two nodes are allowed.

        See also
        --------
        build

        """

        self._graph = graph  # The original graph
        self.graph = None  # The multi directed graph
        self.path = None  # The sorted graph

    def build(self):
        """
        Build graph.

        Returns
        -------
        MultiDiGraph
            A graph object, suitable for computation.

        """

        self.graph = nx.MultiDiGraph()
        nodes = set()
        for node in self._graph["nodes"]:
            node = node.copy()
            id = node.pop("id")
            if id in nodes:
                raise GraphDuplicateNode(f"Duplicate node '{id}'")
            nodes.add(id)
            self.graph.add_node(id, **node)
        if "edges" in self._graph:
            sources = set()
            for edge in self._graph["edges"]:
                src = edge["source"].split(":")
                src_node = src[0]
                if src_node not in nodes:
                    raise GraphUndefinedNode(f"Undefined node '{src_node}'")
                src_port = "o_" + src[1] if len(src) > 1 else "o"
                src = src_node + ":" + src_port
                copy = src in sources
                sources.add(src)
                dst = edge["target"].split(":")
                dst_node = dst[0]
                if dst_node not in nodes:
                    raise GraphUndefinedNode(f"Undefined node '{dst_node}'")
                dst_port = "i_" + dst[1] if len(dst) > 1 else "i"
                self.graph.add_edge(
                    src_node, dst_node, src_port=src_port, dst_port=dst_port, copy=copy
                )
        return self.graph

    def traverse(self):
        """
        Sort nodes in a format suitable for traversal.

        Returns
        -------
        list of dicts:
            List of node ids and their predecessors.

        """

        if not self.graph:
            self.build()
        self.path = []
        for node in nx.algorithms.dag.topological_sort(self.graph):
            predecessors = []
            for predecessor in self.graph.predecessors(node):
                for _, edge in self.graph[predecessor][node].items():
                    predecessors.append(
                        {
                            "node": predecessor,
                            "src_port": edge["src_port"],
                            "dst_port": edge["dst_port"],
                            "copy": edge["copy"],
                        }
                    )
            self.path.append({"node": node, "predecessors": predecessors})
        return self.path



"""Enumerate LSL streams."""

from pylsl import resolve_streams


def enumerate():
    streams = resolve_streams()
    for stream in streams:
        print(
            f"name: {stream.name()}, type: {stream.type()}, source_id: {stream.source_id()}"
        )


if __name__ == "__main__":
    enumerate()



"""Epoching nodes"""

import numpy as np
import pandas as pd
import json
import xarray as xr
from timeflux.core.node import Node
from timeflux.core.exceptions import WorkerInterrupt
from timeflux.helpers.port import match_events


class Epoch(Node):
    """Event-triggered epoching.

    This node continuously buffers a small amount of data (of a duration of ``before`` seconds) from the default input stream.
    When it detects a marker matching the ``event_trigger`` in the ``label`` column of the event input stream, it starts accumulating data for ``after`` seconds.
    It then sends the epoched data to an output stream, and sets the metadata to a dictionary containing the triggering marker and optional event data.
    Multiple, overlapping epochs are authorized. Each concurrent epoch is assigned its own `Port`. For convenience, the first epoch is bound to the default output, so you can avoid enumerating all output ports if you expects only one epoch.

    Attributes:
        i (Port): Default data input, expects DataFrame.
        i_events (Port): Event input, expects DataFrame.
        o (Port): Default output, provides DataFrame and meta.
        o_* (Port): Dynamic outputs, provide DataFrame and meta.

    Args:
        event_trigger (string): The marker name.
        before (float): Length before onset, in seconds.
        after (float): Length after onset, in seconds.

    Example:
        .. literalinclude:: /../examples/epoch.yaml
           :language: yaml

    """

    def __init__(self, event_trigger, before=0.2, after=0.6):

        self._event_trigger = event_trigger
        self._before = pd.Timedelta(seconds=before)
        self._after = pd.Timedelta(seconds=after)
        self._buffer = None
        self._epochs = []

    def update(self):

        # Append to main buffer
        if self.i.data is not None:
            if not self.i.data.empty:
                if self._buffer is None:
                    self._buffer = self.i.data
                else:
                    self._buffer = self._buffer.append(self.i.data)

        # Detect onset
        matches = match_events(self.i_events, self._event_trigger)
        if matches is not None:
            for index, row in matches.iterrows():
                # Start a new epoch
                low = index - self._before
                high = index + self._after
                if self._buffer is not None:
                    if not self._buffer.index.is_monotonic:
                        self.logger.warning(
                            f"Index should be monotonic. Skipping epoch {row['data']}."
                        )
                        return
                    self._epochs.append(
                        {
                            "data": self._buffer[low:high],
                            "meta": {
                                "onset": index,
                                "context": row["data"],
                                "before": self._before.total_seconds(),
                                "after": self._after.total_seconds(),
                            },
                        }
                    )

        # Trim main buffer
        if self._buffer is not None:
            low = self._buffer.index[-1] - self._before
            self._buffer = self._buffer[low:]

        # Update epochs
        if self._epochs and self.i.ready():
            complete = 0
            for epoch in self._epochs:
                high = epoch["meta"]["onset"] + self._after
                last = self.i.data.index[-1]
                if epoch["data"].empty:
                    low = epoch["meta"]["onset"] - self._before
                    mask = (self.i.data.index >= low) & (self.i.data.index <= high)
                else:
                    low = epoch["data"].index[-1]
                    mask = (self.i.data.index > low) & (self.i.data.index <= high)
                # Append
                epoch["data"] = epoch["data"].append(self.i.data[mask])
                # Send if we have enough data
                if last >= high:
                    o = getattr(self, "o_" + str(complete))
                    o.data = epoch["data"]
                    o.meta = {"epoch": epoch["meta"]}
                    complete += 1
            if complete > 0:
                del self._epochs[:complete]  # Unqueue
                self.o = self.o_0  # Bind default output to the first epoch


class ToXArray(Node):
    """Convert multiple epochs to DataArray

    This node iterates over input ports with valid epochs, concatenates them on the
    first axis, and creates a XArray with dimensions ('epoch', 'time', 'space') where
    epoch corresponds to th input ports, time to the ports data index and space to the
    ports data columns.
    A port is considered to be valid if it has meta with key 'epoch' and data with
    expected number of samples.
    If some epoch have an invalid length (which happens when the data has jitter), the
    node either raises a warning, an error or pass.

    Attributes:
        i_* (Port): Dynamic inputs, expects DataFrame and meta.
        o (Port): Default output, provides DataArray and meta.

    Args:
        reporting (string| None): How this function handles epochs with
            invalid length: `warn` will issue a warning with :py:func:`warnings.warn`,
            `error` will raise an exception, `None` will ignore it.
        output (`DataArray`|`Dataset`): Type of output to return
        context_key (string|None): If output type is `Dataset`, key to define the
            target of the event. If `None`, the whole context is considered.

    """

    def __init__(self, reporting="warn", output="DataArray", context_key=None):

        self._reporting = reporting
        self._output = output
        self._context_key = context_key
        self._columns = self._before = self._after = None
        self._ready = False

    def update(self):

        if not self._ready:
            ports_ready = [port for _, _, port in self.iterate("i*") if port.ready()]
            if len(ports_ready) < 1:
                return
            # initialize attributes on first ready port
            port = ports_ready[0]
            if port.ready():
                self._columns = port.data.columns
                self._before = port.meta["epoch"]["before"]
                self._after = port.meta["epoch"]["after"]
                self._num_times = len(port.data)
                self._times = pd.TimedeltaIndex(
                    data=np.linspace(-self._before, self._after, self._num_times),
                    unit="s",
                )
                self._rate = 1 / (self._times[1] - self._times[0]).total_seconds()
                self._ready = True

        ports_ready = [
            port for _, _, port in self.iterate(name="i*") if self._valid_port(port)
        ]

        if not ports_ready:
            return

        list_onset = [port.meta["epoch"].get("onset") for port in ports_ready]
        list_context = [port.meta["epoch"].get("context") for port in ports_ready]
        list_epochs = [port.data for port in ports_ready]

        data = np.stack([epoch.values for epoch in list_epochs], axis=0)

        meta = {
            "epochs_context": list_context,
            "epochs_onset": list_onset,
            "rate": self._rate,
        }

        if self._output == "DataArray":
            if self._context_key is not None:
                data_array = xr.DataArray(
                    data,
                    dims=("target", "time", "space"),
                    coords=(
                        [self._extract_target(context) for context in list_context],
                        self._times,
                        self._columns,
                    ),
                )
            else:
                data_array = xr.DataArray(
                    data,
                    dims=("epoch", "time", "space"),
                    coords=(np.arange(data.shape[0]), self._times, self._columns),
                )
            self.o.data = data_array
            self.o.meta = meta
        else:  # Dataset
            data_array = xr.DataArray(
                data,
                dims=("epoch", "time", "space"),
                coords=(np.arange(data.shape[0]), self._times, self._columns),
            )
            self.o.data = xr.Dataset(
                {
                    "data": data_array,
                    "target": [
                        self._extract_target(context) for context in list_context
                    ],
                }
            )
            self.o.meta = meta

    def _extract_target(self, context):
        if self._context_key is None:
            return context
        else:
            if isinstance(context, str):
                context = json.loads(context)
            return context.get(self._context_key)

    def _valid_port(self, port):
        """ Checks that the port has valid meta and data. """
        if port.data is None or port.data.empty:
            return False
        if "epoch" not in port.meta:
            return False
        if port.data.shape[0] != self._num_times:
            if self._reporting == "error":
                raise WorkerInterrupt(
                    f"Received an epoch with {port.data.shape[0]} "
                    f"samples instead of {self._num_times}."
                )
            elif self._reporting == "warn":
                self.logger.warning(
                    f"Received an epoch with {port.data.shape[0]} "
                    f"samples instead of {self._num_times}. "
                    f"Skipping."
                )
                return False
            else:  # reporting is None
                # be cool
                return False
        return True



"""timeflux.nodes.zmq: a simple 0MQ pub/sub broker"""

import time
import pandas
import zmq
from zmq.devices import ThreadProxy
from timeflux.core.node import Node
from timeflux.core.exceptions import WorkerInterrupt
from timeflux.core.io import Port
import timeflux.core.message


class Broker(Node):

    """Must run in its own graph."""

    def __init__(
        self, address_in="tcp://127.0.0.1:5559", address_out="tcp://127.0.0.1:5560"
    ):
        """
        Initialize frontend and backend.
        If used on a LAN, bind to tcp://*:5559 and tcp://*:5560 instead of localhost.
        """
        try:
            context = zmq.Context.instance()
            self._frontend = context.socket(zmq.XSUB)
            self._frontend.bind(address_in)
            self._backend = context.socket(zmq.XPUB)
            self._backend.bind(address_out)
        except zmq.ZMQError as e:
            self.logger.error(e)

    def update(self):
        """Start a blocking proxy."""
        zmq.proxy(self._frontend, self._backend)


class BrokerMonitored(Node):

    """
    Run a monitored pub/sub proxy.
    Will shut itself down after [timeout] seconds if no data is received.
    Useful for unit testing and replays.
    """

    _last_event = time.time()

    def __init__(
        self,
        address_in="tcp://127.0.0.1:5559",
        address_out="tcp://127.0.0.1:5560",
        timeout=5,
    ):

        self._timeout = timeout

        try:

            # Capture
            address_monitor = "inproc://monitor"
            context = zmq.Context.instance()
            self._monitor = context.socket(zmq.PULL)
            self._monitor.bind(address_monitor)

            # Proxy
            proxy = ThreadProxy(zmq.XSUB, zmq.XPUB, zmq.PUSH)
            proxy.bind_in(address_in)
            proxy.bind_out(address_out)
            proxy.connect_mon(address_monitor)
            # proxy.setsockopt_mon(zmq.CONFLATE, True) # Do not clutter the network
            proxy.start()

        except zmq.ZMQError as error:
            self.logger.error(error)

    def update(self):
        """Monitor proxy"""
        if self._timeout == 0:
            pass
        now = time.time()
        count = 0
        try:
            while True:
                self._monitor.recv_multipart(zmq.NOBLOCK, copy=False)
                self._last_event = now
                count += 1
        except zmq.ZMQError:
            if count > 0:
                self.logger.debug("Received %d messages", count)
            if (now - self._last_event) > self._timeout:
                raise WorkerInterrupt("No data after %d seconds" % self._timeout)


class BrokerLVC(Node):
    """A monitored pub/sub broker with last value caching."""

    def __init__(
        self,
        address_in="tcp://127.0.0.1:5559",
        address_out="tcp://127.0.0.1:5560",
        timeout=1000,
    ):
        self._timeout = timeout
        try:
            context = zmq.Context.instance()
            self._frontend = context.socket(zmq.SUB)
            self._frontend.setsockopt(zmq.SUBSCRIBE, b"")
            self._frontend.bind(address_in)
            self._backend = context.socket(zmq.XPUB)
            self._backend.setsockopt(zmq.XPUB_VERBOSE, True)
            self._backend.bind(address_out)
            self._poller = zmq.Poller()
            self._poller.register(self._frontend, zmq.POLLIN)
            self._poller.register(self._backend, zmq.POLLIN)
        except zmq.ZMQError as error:
            self.logger.error(error)

    def update(self):
        """Main poll loop."""
        cache = {}
        while True:
            events = dict(self._poller.poll(self._timeout))
            # Any new topic data we cache and then forward
            if self._frontend in events:
                message = self._frontend.recv_multipart()
                topic, current = message
                cache[topic] = current
                self._backend.send_multipart(message)
            # handle subscriptions
            # When we get a new subscription we pull data from the cache:
            if self._backend in events:
                event = self._backend.recv()
                # print(event)
                # Event is one byte 0=unsub or 1=sub, followed by topic
                # if event[0] == b'\x01':
                if event[0] == 1:
                    topic = event[1:]
                    if topic in cache:
                        self.logger.debug(
                            "Sending cached topic %s", topic.decode("utf-8")
                        )
                        self._backend.send_multipart([topic, cache[topic]])


class Pub(Node):
    def __init__(
        self, topic, address="tcp://127.0.0.1:5559", serializer="pickle", wait=0
    ):
        """Create a publisher"""
        self._topic = topic.encode("utf-8")
        self._serializer = getattr(timeflux.core.message, serializer + "_serialize")
        try:
            context = zmq.Context.instance()
            self._socket = context.socket(zmq.PUB)
            self._socket.setsockopt(zmq.LINGER, 0)
            self._socket.connect(address)
        except zmq.ZMQError as e:
            self.logger.error(e)

        # Quick fix to the slow joiner syndrome
        # TODO: remove when Last Value Caching is implemented
        # Wait for subscribers to connect
        # http://zguide.zeromq.org/page%3aall#Getting-the-Message-Out
        # http://zguide.zeromq.org/page%3aall#Node-Coordination
        # http://zguide.zeromq.org/page%3aall#Last-Value-Caching
        # https://stackoverflow.com/questions/30864145/zmq-no-subscription-message-on-xpub-socket-for-multiple-subscribers-last-value
        time.sleep(wait)

    def update(self):
        for name, suffix, port in self.iterate("i*"):
            if port.ready() or port.meta:
                if not suffix:
                    topic = self._topic
                else:
                    topic = self._topic + suffix.encode("utf-8")
                try:
                    self._socket.send_serialized(
                        [topic, port.data, port.meta], self._serializer
                    )
                except zmq.ZMQError as e:
                    self.logger.error(e)


class Sub(Node):
    def __init__(
        self, topics=[""], address="tcp://127.0.0.1:5560", deserializer="pickle"
    ):
        """Create a subscriber"""
        try:
            context = zmq.Context.instance()
            self._socket = context.socket(zmq.SUB)
            self._socket.connect(address)
            for topic in topics:
                self._socket.setsockopt(zmq.SUBSCRIBE, topic.encode("utf-8"))
                if topic:
                    if not topic.isidentifier():
                        raise ValueError("Invalid topic name: %s" % topic)
            self._deserializer = getattr(
                timeflux.core.message, deserializer + "_deserialize"
            )
        except zmq.ZMQError as e:
            self.logger.error(e)

    def update(self):
        self._chunks = {}
        try:
            while True:
                [topic, data, meta] = self._socket.recv_serialized(
                    self._deserializer, zmq.NOBLOCK
                )
                if not topic in self._chunks:
                    self._chunks[topic] = {"data": [], "meta": {}}
                self._append_data(topic, data)
                self._append_meta(topic, meta)
        except zmq.ZMQError:
            pass  # No more data
        self._update_ports()

    def _append_data(self, topic, data):
        if data is not None:
            self._chunks[topic]["data"].append(data)

    def _append_meta(self, topic, meta):
        if meta:
            self._chunks[topic]["meta"].update(meta)

    def _update_ports(self):
        for topic in self._chunks.keys():
            if len(self._chunks[topic]["data"]) == 0:
                data = None
            elif len(self._chunks[topic]["data"]) == 1:
                data = self._chunks[topic]["data"][0]
            else:
                data = pandas.concat(self._chunks[topic]["data"])
            meta = self._chunks[topic]["meta"]
            self._update_port(topic, data, meta)

    def _update_port(self, topic, data, meta):
        getattr(self, "o_" + topic).data = data
        getattr(self, "o_" + topic).meta = meta



import os
import sys
import threading
import time
import logging
import logging.config
import logging.handlers
import coloredlogs
from multiprocessing import Queue
from datetime import datetime, timezone


_QUEUE = None


class _UTCFormatterConsole(coloredlogs.ColoredFormatter):
    converter = time.gmtime


class _UTCFormatterFile(logging.Formatter):
    converter = time.gmtime


class Handler:
    def __init__(self):
        self.logger = logging.getLogger("timeflux")

    def handle(self, record):
        self.logger.handle(record)


def init_listener(level_console="INFO", level_file="DEBUG", file=None):

    q = get_queue()

    level_styles = {
        "debug": {"color": "white"},
        "info": {"color": "cyan"},
        "warning": {"color": "yellow"},
        "error": {"color": "red"},
        "critical": {"color": "magenta"},
    }
    field_styles = {
        "asctime": {"color": "blue"},
        "levelname": {"color": "black", "bright": True},
        "processName": {"color": "green"},
    }
    record_format = "%(asctime)s %(levelname)-10s %(module)-12s %(process)-8s %(processName)-16s %(message)s"

    # On Windows, the colors will not work unless we initialize the console with colorama
    if sys.platform.startswith("win"):
        try:
            import colorama

            colorama.init()
        except ImportError:
            # On Windows, without colorama, do not use the colors
            level_styles = {}
            field_styles = {}

    # Define config
    config = {
        "version": 1,
        "root": {
            "handlers": ["default"],
        },
        "loggers": {
            "timeflux": {"propagate": False, "level": "DEBUG", "handlers": ["console"]},
        },
        "formatters": {
            "console": {
                "()": "timeflux.core.logging._UTCFormatterConsole",
                "format": record_format,
                "datefmt": "%Y-%m-%d %H:%M:%S,%f",
                "field_styles": field_styles,
                "level_styles": level_styles,
            },
            "file": {
                "class": "timeflux.core.logging._UTCFormatterFile",
                "format": record_format,
            },
        },
        "handlers": {
            "default": {
                "class": "logging.StreamHandler",
                "formatter": "console",
                "stream": "ext://sys.stdout",
            },
            "console": {
                "level": level_console,
                "class": "logging.StreamHandler",
                "formatter": "console",
                "stream": "ext://sys.stdout",
            },
            "file": {
                "level": level_file,
                "class": "logging.FileHandler",
                "filename": "timeflux.log",
                "mode": "a",
                "delay": True,
                "formatter": "file",
            },
        },
    }

    if file:
        now = datetime.now(timezone.utc)
        config["loggers"]["timeflux"]["handlers"].append("file")
        config["handlers"]["file"]["filename"] = now.strftime(file)

    logging.config.dictConfig(config)

    queue = get_queue()
    listener = logging.handlers.QueueListener(queue, Handler())
    listener.start()


def terminate_listener():
    if _QUEUE:
        _QUEUE.put_nowait(None)


def init_worker(queue):

    config = {
        "version": 1,
        "handlers": {
            "queue": {"class": "logging.handlers.QueueHandler", "queue": queue},
        },
        "loggers": {
            "timeflux": {"propagate": False, "level": "DEBUG", "handlers": ["queue"]}
        },
        "root": {"handlers": ["queue"]},
    }

    logging.config.dictConfig(config)


def get_queue():
    if not _QUEUE:
        globals()["_QUEUE"] = Queue()
    return _QUEUE



"""timeflux.nodes.random: generate random data"""

import random
import numpy as np
from timeflux.core.node import Node


class Random(Node):
    def __init__(
        self,
        columns=5,
        rows_min=2,
        rows_max=10,
        value_min=0,
        value_max=9,
        names=None,
        seed=None,
    ):
        """Return random integers from value_min to value_max (inclusive)"""
        self._rows_min = rows_min
        self._rows_max = rows_max
        self._value_min = value_min
        self._value_max = value_max
        self._names = names
        self._columns = len(names) if names else columns
        random.seed(seed)
        np.random.seed(seed)

    def update(self):
        rows = random.randint(self._rows_min, self._rows_max)
        shape = (rows, self._columns)
        self.o.set(
            np.random.randint(self._value_min, self._value_max + 1, size=shape),
            names=self._names,
        )



"""timeflux.core.manager: manage workers"""

import logging
import sys
import os
import time
import signal
import json
import yaml
from jinja2 import Template
from timeflux.core.validate import validate
from timeflux.core.worker import Worker


class Manager:

    """Load configuration and spawn workers."""

    def __init__(self, config):
        """Load configuration

        Args:
            config (string|dict): The configuration can either be a path to a YAML or JSON file or a dict.

        """

        # Initialize logger
        self.logger = logging.getLogger(__name__)

        # Hold the names of the imported applications
        self._imports = []

        # Hold the graphs
        self._graphs = []

        # Hold the children processes
        self._processes = []

        # Load application
        if isinstance(config, dict):
            app = config
            path = os.getcwd()
        elif isinstance(config, str):
            app = self._load_file(config)
            self._imports.append(os.path.abspath(config))
            path = os.path.dirname(self._imports[0])
        else:
            raise ValueError("Could not load application file")

        # Validate
        validate(app)

        # Populate the graph list
        if "graphs" in app:
            self._graphs = app["graphs"]

        # Import sub applications
        self._import(app, path)

        # Update the search path
        for app in self._imports:
            path = os.path.dirname(app)
            if path not in sys.path:
                sys.path.append(path)

    def run(self):
        """Launch as many workers as there are graphs."""
        try:
            # Launch workers
            self._launch()
            # Monitor them
            self._monitor()
        except KeyboardInterrupt:
            # Ignore further interrupts
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            self.logger.info("Interrupting")
        # Terminate gracefully
        self._terminate()

    def _launch(self):
        """Launch workers."""
        for graph in self._graphs:
            worker = Worker(graph)
            process = worker.run()
            self._processes.append(process)
            self.logger.debug("Worker spawned with PID %d", process.pid)

    def _monitor(self):
        """Wait for at least one worker to terminate."""
        if not self._processes:
            return
        while True:
            if any(not process.is_alive() for process in self._processes):
                return
            time.sleep(0.1)

    def _terminate(self):
        """Terminate all workers."""
        # https://bugs.python.org/issue26350
        interrupt = signal.CTRL_C_EVENT if sys.platform == "win32" else signal.SIGINT
        # Try to terminate gracefully
        for process in self._processes:
            if process.is_alive():
                os.kill(process.pid, interrupt)
        # Wait 10 seconds and kill the remaining ones
        self._wait(10)
        for process in self._processes:
            if process.is_alive():
                process.terminate()

    def _wait(self, timeout=None):
        """Wait for all workers to die."""
        if not self._processes:
            return
        start = time.time()
        while True:
            try:
                if all(not process.is_alive() for process in self._processes):
                    # All the workers are dead
                    return
                if timeout and time.time() - start >= timeout:
                    # Timeout
                    return
                time.sleep(0.1)
            except:
                pass

    def _import(self, app, path):
        if not "import" in app:
            return
        old_path = os.getcwd()
        os.chdir(path)
        for filename in app["import"]:
            filename = os.path.abspath(filename)
            if filename in self._imports:
                self.logger.debug("Application %s will not be loaded twice", filename)
                continue
            self.logger.debug("Importing %s", filename)
            self._imports.append(filename)
            sub = self._load_file(filename)
            try:
                validate(sub)
            except ValueError as error:
                raise ValueError(f"Validation failed ({filename})")
            if "graphs" in sub:
                self._graphs += sub["graphs"]
            self._import(sub, os.path.dirname(filename))
        os.chdir(old_path)

    def _load_file(self, filename):

        # Read file as string
        with open(filename) as stream:
            app = stream.read()

        # Parse the template
        app = self._parse_template(app)

        # Return a dict
        extension = filename.split(".")[-1]
        if extension in ("yml", "yaml"):
            return yaml.safe_load(app)
        elif extension == "json":
            return json.loads(app)

    def _parse_template(self, template):
        template = Template(template)
        return template.render(dict(os.environ))



import pytest
from time import sleep
from timeflux.helpers.background import Task

class DummyWorker():
    def echo(self, message='hello', delay=0, fail=False):
        sleep(delay)
        if fail: raise Exception('failed')
        self.message = message
        return(self.message)

def test_default(working_path):
    task = Task(DummyWorker(), 'echo').start()
    while not task.done:
        status = task.status()
    assert status['result'] == 'hello'
    assert status['instance'].message == 'hello'

def test_args(working_path):
    task = Task(DummyWorker(), 'echo', 'foobar').start()
    while not task.done:
        status = task.status()
    assert status['result'] == 'foobar'

def test_kwargs(working_path):
    task = Task(DummyWorker(), 'echo', message='foobar').start()
    while not task.done:
        status = task.status()
    assert status['result'] == 'foobar'

def test_exception(working_path):
    task = Task(DummyWorker(), 'echo', fail=True).start()
    while not task.done:
        status = task.status()
    assert status['success'] == False
    assert status['exception'].args[0] == 'failed'



import argparse
import os, sys
import time
import random

path = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, path  + '/../../')

from timeflux.core.sync import Server, Client

parser = argparse.ArgumentParser()
parser.add_argument('--host', help='The server address', default='')
parser.add_argument('--port', help='The server port', type=int, default=12300)
parser.add_argument('mode', help='The mode: "server" or "client".')
args = parser.parse_args()

def dummy_time(offset=1):
    time.sleep(random.random() / 2) # simulate network latency
    return time.perf_counter() + offset

if args.mode == 'server':
    server = Server(args.host, args.port, now=dummy_time)
    try:
        server.start()
    except KeyboardInterrupt:
        server.stop()

if args.mode == 'client':
    client = Client(args.host, args.port, rounds=100)
    try:
        client.sync()
    except KeyboardInterrupt:
        client.stop()




"""Test configuration"""

import os
import sys
import importlib
import pytest
import numpy as np
import random as rand

def pytest_configure(config):
    pytest.path = os.path.dirname(os.path.abspath(__file__))

@pytest.fixture()
def random():
    """Initialize the random seeds for reproducibility."""
    rand.seed(0)
    np.random.seed(0)

@pytest.fixture()
def working_path(request):
    """Change the working directory to the test file path."""
    old = os.getcwd()
    new = os.path.dirname(request.fspath)
    os.chdir(new)
    yield
    os.chdir(old)


import numpy as np
import pandas as pd
import xarray as xr
from timeflux.helpers.mne import mne_to_xarray, xarray_to_mne

rate = 10.0  # Sampling rate
# Generate epochs data, ie. random data in shape (n_epochs, n_channels, n_times)
times = pd.TimedeltaIndex(np.arange(-.2, .8, 1 / rate), unit='s')
n_epochs = 3
n_channels = 5
n_times = len(times)

channels = [f'ch{k}' for k in range(n_channels)]
# Generate data and meta as it were from output port of node eEpochsToDataArray
original_data = xr.DataArray(np.random.rand(n_epochs, n_times, n_channels),
                             dims=('epoch', 'time', 'space'),
                             coords=(np.arange(n_epochs), times, channels))
original_meta = {'epochs_context': [{'target': 'foo'}] + [{'target': 'bar'}] * (n_epochs - 1),
                 'rate': rate,
                 'epochs_onset': list(pd.date_range(
                     start='2018-01-01',
                     periods=n_epochs,
                     freq=pd.DateOffset(seconds=1)))}
# context_key is the key to select the label in the context
context_key = 'target'
event_id = {'foo': 0, 'bar': 1}


def test_mne_dataarray_conversions():
    epochs = xarray_to_mne(original_data, original_meta, context_key, event_id)
    after_conversion_data, after_conversion_meta = mne_to_xarray(epochs, context_key, event_id)
    # assert that the passage in mne keeps the data unchanged
    xr.testing.assert_equal(after_conversion_data, original_data)
    assert original_meta == after_conversion_meta



"""Sliding windows"""

from timeflux.core.node import Node
import pandas as pd


class Window(Node):

    """Provide sliding windows.

    Attributes:
        i (Port): Default data input, expects DataFrame.

    Args:
        length (float|int): The length of the window, in seconds or samples.
        step (float|int|None): The minimal sliding step, in seconds or samples.
            If None (the default), the step will be set to the length of the window.
            If 0, the data will be sent as soon as it is available.
        index (string):
            If "time" (the default), the length of the window is in seconds.
            If "sample", the length of the window is in samples.
        epochs (boolean):
            Whether the defaut output should be bound to an numbered output, thus
            simulating an epoch. This could be useful if piped to a Machine Learning
            node.

    """

    def __init__(self, length, step=None, index="time", epochs=False):

        if index not in ["time", "sample"]:
            raise ValueError("Invalid `index` value.")
        mixin = TimeWindow if index == "time" else SampleWindow
        self.__class__ = type("SlidingWindow", (mixin, Window), {})
        self.__init__(length, step)
        if epochs:
            self.bind("o", "o_0")

    def update(self):

        pass


class TimeWindow(Node):
    def __init__(self, length, step=None):

        if step is None:
            step = length
        if step > length:
            raise ValueError("`step` must be less than or equal to `length`.")
        self._length = pd.Timedelta(seconds=length)
        self._step = pd.Timedelta(seconds=step)
        self._buffer = None

    def update(self):

        # Return immediately if we don't have any data
        if not self.i.ready():
            return

        # Sanity check
        if not self.i.data.index.is_monotonic:
            self.logger.warning("Indices are non-monotonic.")

        # Append new data
        if self._buffer is None:
            self._buffer = self.i.data
        else:
            self._buffer = self._buffer.append(self.i.data)

        # Update the default output if we have enough data
        low = self._buffer.index[0]
        high = low + self._length
        if self._buffer.index[-1] >= high:
            self.o.data = self._buffer[self._buffer.index < high]
            self.o.meta = self.i.meta
            self._buffer = self._buffer[self._buffer.index >= low + self._step]

        # Make sure we are not overflowing
        if (
            not self._buffer.empty
            and (self._buffer.index[-1] - self._buffer.index[0]) > self._length
        ):
            if self._step is not 0:
                self.logger.warning(
                    "This node is falling behind: it is receiving "
                    "more data than it can send. Check the window "
                    "parameters and the graph rate."
                )
            self._buffer = self._buffer[
                self._buffer.index > self._buffer.index[-1] - self._length + self._step
            ]


class SampleWindow(Node):
    def __init__(self, length, step=None):

        if step is None:
            step = length
        self._length = length
        self._step = step
        self._buffer = None

    def update(self):

        if not self.i.ready():
            return

        # Append new data
        if self._buffer is None:
            self._buffer = self.i.data
        else:
            self._buffer = self._buffer.append(self.i.data)

        # Make sure we have enough data
        if len(self._buffer) >= self._length:
            self.o.data = self._buffer[-self._length :]
            self.o.meta = self.i.meta
            # Step
            self._buffer = self.o.data[self._step :]



"""Tests for gate.py"""
import numpy as np
import pandas as pd
import xarray as xr
from timeflux.helpers.testing import DummyData, DummyXArray
from timeflux.nodes.gate import Gate

xarray_data = DummyXArray()
pandas_data = DummyData()


node = Gate(event_opens='foo_begins', event_closes='foo_ends', truncate=True)



def test_gate_silent():
    pandas_data.reset()
    # Send data but no event
    node.i.data = pandas_data.next(20)
    node.update()

    assert node._status == 'closed'
    assert node.o.data == None


def test_send_opening_closing_event_in_separate_chunks():
    pandas_data.reset()
    # Send an opening event
    node.clear()
    node.i.data = pandas_data.next(5)
    time_open = pd.Timestamp('2018-01-01 00:00:00.104507143')  # Sync event to second sample
    event = pd.DataFrame([['foo_begins']], [time_open], columns=['label'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    expected_data = pd.DataFrame(
        [
            [0.658783, 0.692277, 0.849196, 0.249668, 0.489425],
            [0.221209, 0.987668, 0.944059, 0.039427, 0.705575],
            [0.925248, 0.180575, 0.567945, 0.915488, 0.033946],
            [0.69742, 0.297349, 0.924396, 0.971058, 0.944266],
        ],
        [
            pd.Timestamp('2018-01-01 00:00:00.104507143'),
            pd.Timestamp('2018-01-01 00:00:00.202319939'),
            pd.Timestamp('2018-01-01 00:00:00.300986584'),
            pd.Timestamp('2018-01-01 00:00:00.396560186')

        ]
    )

    expected_meta = {'gate_status': 'open'}

    assert node._status == 'open'
    assert node._trigger == 'foo_ends'
    assert node.o.meta == expected_meta
    pd.testing.assert_frame_equal(node.o.data, expected_data)

    # Send a closing event
    node.clear()
    node.i.data = pandas_data.next(5)
    time_close = pd.Timestamp('2018-01-01 00:00:00.595580836')  # Sync event to second sample
    event = pd.DataFrame([['foo_ends']], [time_close], columns=['label'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    expected_data = pd.DataFrame(
        [
            [0.474214, 0.862043, 0.844549, 0.3191, 0.828915],
            [0.037008, 0.59627, 0.230009, 0.120567, 0.076953]
        ],
        [
            pd.Timestamp('2018-01-01 00:00:00.496559945'),
            pd.Timestamp('2018-01-01 00:00:00.595580836')

        ]
    )
    expected_meta = {'gate_status': 'closed',
                     'gate_times': [time_open, time_close]}


    assert node._status == 'closed'
    assert node._trigger == 'foo_begins'
    assert node.o.meta == expected_meta
    pd.testing.assert_frame_equal(node.o.data, expected_data)


def test_send_opening_closing_event_in_same_chunk():
    # Send an opening event and a closing event in the same chunk
    pandas_data.reset()
    node.clear()
    node.i.data = pandas_data.next(5)
    time_open = pd.Timestamp('2018-01-01 00:00:00.1')
    time_close = pd.Timestamp('2018-01-01 00:00:00.3')
    event = pd.DataFrame([['foo_begins'], ['foo_ends']], [time_open, time_close],
                         columns=['label'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    expected_data = pd.DataFrame(
        [
            [0.658783, 0.692277, 0.849196, 0.249668, 0.489425],
            [0.221209, 0.987668, 0.944059, 0.039427, 0.705575]
        ],
        [
            pd.Timestamp('2018-01-01 00:00:00.104507143'),
            pd.Timestamp('2018-01-01 00:00:00.202319939')
        ]
    )
    expected_meta = {'gate_status': 'closed',
                     'gate_times': [time_open, time_close]}

    assert node.o.meta == expected_meta
    pd.testing.assert_frame_equal(node.o.data, expected_data)


    assert node._status == 'closed'
    assert node._trigger == 'foo_begins'


def test_xarray_data():
    # Send an opening event and a closing event in the same chunk with XArray data
    xarray_data.reset()
    node.clear()
    node.i.data = xarray_data.next(5)
    time_open = pd.Timestamp('2018-01-01 00:00:00.1')
    time_close = pd.Timestamp('2018-01-01 00:00:00.3')
    event = pd.DataFrame([['foo_begins'], ['foo_ends']], [time_open, time_close],
                         columns=['label'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    expected_data = xr.DataArray(data=
    [
        [0.658783, 0.692277, 0.849196, 0.249668, 0.489425],
        [0.221209, 0.987668, 0.944059, 0.039427, 0.705575]
    ],
        coords=[
            [
                pd.Timestamp('2018-01-01 00:00:00.104507143'),
                pd.Timestamp('2018-01-01 00:00:00.202319939')
            ],
            np.arange(5)],
        dims=['time', 'space'])

    expected_meta = {'gate_status': 'closed',
                     'gate_times': [time_open, time_close]}

    assert node.o.meta == expected_meta
    xr.testing.assert_equal(node.o.data, expected_data)

    assert node._status == 'closed'
    assert node._trigger == 'foo_begins'



"""Simple debugging nodes"""

import csv
import pandas as pd
from timeflux.core.node import Node


class Display(Node):
    """Display input."""

    def __init__(self, meta=False, data=True):
        self._meta = meta
        self._data = data

    def update(self):
        if self.i.ready() and self._data:
            self.logger.debug("\n %s" % self.i.data)
        if self.i.meta and self._meta:
            self.logger.debug("\n %s" % self.i.meta)


class Dump(Node):
    """Dump to CSV."""

    def __init__(self, fname="/tmp/dump.csv"):
        self.writer = csv.writer(open(fname, "a"))

    def update(self):
        if self.i.ready():
            self.i.data["index"] = self.i.data.index.values.astype("float64") / 1e9
            self.writer.writerows(self.i.data.values.tolist())



"""Timeflux entry point"""

import sys
import os
import logging
from argparse import ArgumentParser
from runpy import run_module
from dotenv import load_dotenv
from timeflux import __version__
from timeflux.core.logging import init_listener, terminate_listener
from timeflux.core.manager import Manager

LOGGER = logging.getLogger(__name__)


def main():
    sys.path.append(os.getcwd())
    args = _args()
    _init_env(args.env_file, args.env)
    _init_logging(args.debug)
    LOGGER.info("Timeflux %s" % __version__)
    _run_hook("pre")
    try:
        Manager(args.app).run()
    except Exception as error:
        LOGGER.error(error)
    _terminate()


def _args():
    parser = ArgumentParser()
    parser.add_argument(
        "-v", "--version", action="version", version="Timeflux %s" % __version__
    )
    parser.add_argument(
        "-d",
        "--debug",
        default=False,
        action="store_true",
        help="enable debug messages",
    )
    parser.add_argument(
        "-E", "--env-file", default="./.env", help="path to an environment file"
    )
    parser.add_argument("-e", "--env", action="append", help="environment variables")
    parser.add_argument("app", help="path to the YAML or JSON application file")
    args = parser.parse_args()
    return args


def _terminate():
    _run_hook("post")
    LOGGER.info("Terminated")
    terminate_listener()
    sys.exit(0)


def _init_env(file, vars):
    load_dotenv(file)
    if vars is not None:
        for env in vars:
            if "=" in env:
                env = env.split("=", 1)
                os.environ[env[0]] = env[1]


def _init_logging(debug):
    level_console = (
        "DEBUG" if debug else os.getenv("TIMEFLUX_LOG_LEVEL_CONSOLE", "INFO")
    )
    level_file = os.getenv("TIMEFLUX_LOG_LEVEL_FILE", "DEBUG")
    file = os.getenv("TIMEFLUX_LOG_FILE", None)
    init_listener(level_console, level_file, file)


def _run_hook(name):
    module = os.getenv("TIMEFLUX_HOOK_" + name.upper())
    if module:
        LOGGER.info("Running %s hook" % name)
        try:
            run_module(module)
        except ImportError as error:
            LOGGER.error(error)
        except Exception as error:
            LOGGER.exception(error)



"""Tests for miscellaneous nodes"""

import pytest
import pandas as pd
import numpy as np
from pandas.testing import assert_frame_equal
from timeflux.core.registry import Registry
from timeflux.nodes.random import Random
from timeflux.nodes.hdf5 import Replay
from timeflux.helpers.clock import effective_rate

Registry.cycle_start = 0
Registry.rate = 1

def test_random():
    node = Random(2, 2, 2, 0, 10, seed=1)
    node.update()
    expected_index = np.empty(2, dtype='datetime64[us]')
    expected_data = [[5,8],[9,5]]
    expected = pd.DataFrame(expected_data, index=expected_index)
    assert_frame_equal(node.o.data, expected)

# TODO: make a generic test for replay
#def test_replay():
    # filename = pytest.path + '/data/replay.hdf5'
    # keys = ['/nexus/signal/nexus_signal_raw', '/unity/events/unity_events']
    # node = Replay(filename, keys, 40)
    # node.update()
    # rate = effective_rate(node.o_nexus_signal_nexus_signal_raw.data)
    # assert True


"""Monitor a signal"""

from timeflux.core.node import Node


class Monitor(Node):
    def update(self):
        status = 0.0
        if self.i.data is not None:
            if not self.i.data.empty:
                status = 1.0
        self.o.set([status], names=["streaming"])



"""timeflux.nodes.osc: Simple OSC client and server"""

import pandas as pd
from threading import Thread, Lock
from pythonosc.dispatcher import Dispatcher
from pythonosc.osc_server import BlockingOSCUDPServer
from pythonosc.udp_client import SimpleUDPClient
from timeflux.helpers.clock import now
from timeflux.core.node import Node


class Server(Node):

    """A simple OSC server."""

    def __init__(self, addresses=[], ip="127.0.0.1", port=5005):
        self._server = None
        self._data = {}
        self._lock = Lock()
        if not addresses or not isinstance(addresses, list):
            raise ValueError("You must provide a list of addresses.")
        dispatcher = Dispatcher()
        for address in addresses:
            self._data[self._address_to_port(address)] = {"timestamps": [], "rows": []}
            dispatcher.map(address, self._handler)
        self._server = BlockingOSCUDPServer((ip, port), dispatcher)
        Thread(target=self._server.serve_forever).start()

    def update(self):
        with self._lock:
            for port, data in self._data.items():
                if data["rows"]:
                    getattr(self, port).set(data["rows"], data["timestamps"])
                    self._data[port] = {"timestamps": [], "rows": []}

    def terminate(self):
        if self._server:
            self._server.shutdown()

    def _handler(self, address, *args):
        time = now()
        port = self._address_to_port(address)
        values = list(args)
        with self._lock:
            self._data[port]["rows"].append(values)
            self._data[port]["timestamps"].append(time)

    def _address_to_port(self, address):
        address = "/" + address if not address.startswith("/") else address
        return "o" + address.replace("/", "_")


class Client(Node):

    """A simple OSC client."""

    def __init__(self, address="", ip="127.0.0.1", port=5005):
        if not address or not isinstance(address, str):
            raise ValueError("You must provide an address.")
        self._address = address
        self._client = SimpleUDPClient(ip, port)

    def update(self):
        if self.i.data is not None:
            for row in self.i.data.itertuples(index=False):
                self._client.send_message(self._address, row)



"""Accumulation nodes that either, stack, append or, concatenate data after a gate"""

import pandas as pd
import xarray as xr
from timeflux.core.node import Node


class AppendDataFrame(Node):
    """Accumulates and appends data of type DataFrame after a gate.

    This node should be plugged after a Gate. As long as it receives data,
    it appends them to an internal buffer. When it receives a meta with key
    `gate_status` set to `closed`, it releases the accumulated data and empty the
    buffer.

    Attributes:
        i (Port): Default data input, expects DataFrame and meta
        o (Port): Default output, provides DataFrame

    Args:
        **kwargs: key word arguments to pass to pandas.DataFrame.append method.

    """

    def __init__(self, meta_keys=None, **kwargs):

        super().__init__()
        self._meta_keys = meta_keys
        self._kwargs = kwargs
        self._reset()

    def _reset(self):
        self._data = pd.DataFrame()
        self._meta = []

    def _release(self):
        self.logger.info(
            f"AppendDataFrame is releasing {len(self._data)} " f"accumulated rows."
        )
        self.o.data = self._data
        if self._meta_keys is None:
            self.o.meta = {"accumulate": self._meta}
        else:
            self.o.meta = {key: [] for key in self._meta_keys}
            for meta_key in self._meta_keys:
                for meta in self._meta:
                    self.o.meta[meta_key] += meta.get(meta_key, [])

    def update(self):

        gate_status = self.i.meta.get("gate_status")

        if self.i.ready():
            # update the meta
            self._meta.append(self.i.meta)

            # append the data
            self._data = self._data.append(self.i.data, **self._kwargs)

        # if gate is close, release the data and reset the buffer
        if gate_status == "closed" and not self._data.empty:
            self._release()
            self._reset()


class AppendDataArray(Node):
    """Accumulates and appends data of type XArray after a gate.

    This node should be plugged after a Gate. As long as it receives DataArrays,
    it appends them to a buffer list. When it receives a meta with key
    `gate_status` set to `closed`, it concatenates the list of accumulated DataArray,
    releases it and empty the buffer list.

    Attributes:
        i (Port): Default data input, expects DataArray and meta
        o (Port): Default output, provides DataArray

    Args:
        dim: Name of the dimension to concatenate along.
        **kwargs: key word arguments to pass to xarray.concat method.

    """

    def __init__(self, dim, meta_keys=None, **kwargs):

        super().__init__()
        self._dim = dim
        self._meta_keys = meta_keys
        self._kwargs = kwargs
        self._reset()

    def _reset(self):
        self._data_list = []
        self._meta = []

    def _release(self):
        self.logger.info(
            f"AppendDataArray is releasing {len(self._data_list)} "
            f"accumulated data chunks."
        )
        self.o.data = xr.concat(self._data_list, self._dim, **self._kwargs)
        if self._meta_keys is None:
            self.o.meta = {"accumulate": self._meta}
        else:
            self.o.meta = {key: [] for key in self._meta_keys}
            for meta_key in self._meta_keys:
                for meta in self._meta:
                    self.o.meta[meta_key] += meta.get(meta_key, [])

    def update(self):

        gate_status = self.i.meta.get("gate_status")

        # append the data
        if self.i.ready():
            self._data_list.append(self.i.data)
            # update the meta
            self._meta.append(self.i.meta)

        # if gate is close, release the data and reset the buffer
        if gate_status == "closed" and self._data_list:
            self._release()
            self._reset()
